
;;; rrse.el --- Ruby Reference Searcher for Emacsen

;; Copyright (C) 2017 by Ippei Obayashi

;; Author: Ippei Obayashi <ohai@kmc.gr.jp>
;; URL: http://www.kmc.gr.jp/~ohai/rrse.html
;; Version: 0.7
;; Package-Requires: ()

;; This software is distributed under GNU general Public Licence version 2 or later.
;; 
;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;;; Code:

(provide 'rrse)

;;;; User customizable varaible

(defcustom rrse-idle-delay 0.5 "RRSE delay seconds")
(defcustom rrse-ruby-mode 'ruby-mode "Major mode for RRSE")
(defcustom rrse-ruby-mode-hook 'ruby-mode-hook "Major mode hook for RRSE")
(defcustom rrse-directory "~/.rrse" "RRSE directory")

;;;; Internal variables
(defvar rrse-enable t)
(defvar rrse-program-name "rrse")
(defvar rrse-process-buffer nil)
(defvar rrse-process nil)
(defvar rrse-query-timeout 5)

;;;; Functions
(defun rrse-get-current-word ()
  "Return the word at the point"
  (save-excursion
    (buffer-substring-no-properties
     (progn (skip-syntax-backward "w_") (point))
     (progn (skip-syntax-forward "w_") (point)))))

(defun rrse-method-call-p ()
  "Return true if the current word is a method call by using dot"
  (save-excursion
    (skip-syntax-backward "w_")
    (skip-syntax-backward " ")
    (eq (char-before) ?.)))

(defun rrse-get-receiver ()
  "Return the receiver of the method call at the cursor"
  (when (rrse-method-call-p)
    (save-excursion
      (skip-syntax-backward "w_")
      (skip-syntax-backward " ")
      (backward-char)
      (let ((receiver-end-point (point)))
        (skip-syntax-backward "w_")
        (buffer-substring-no-properties (point) receiver-end-point)))))

(defun rrse-start-process (buffer)
  "Invoke \"rrse pipe-query\""
  (start-process "rrse" buffer
                 rrse-program-name "-d" (expand-file-name rrse-directory)
                 "pipe-query" "-s" " "))

(defun rrse-init-process ()
  "Invoke \"rrse pipe-query\" if rrse-process is nil and set rrse-process"
  (when (not rrse-process)
    (setq rrse-process-buffer (get-buffer-create " *rrse*"))
    (setq rrse-process (rrse-start-process rrse-process-buffer))
    (set-process-coding-system rrse-process 'utf-8-unix 'utf-8-unix)
    (set-process-query-on-exit-flag rrse-process nil)))

(defun rrse-kill-process ()
  "Kill rrse pipe-query process if the process is alive"
  (when (and rrse-process (eq (process-status rrse-process) 'run))
    (kill-process rrse-process)
    (setq rrse-process nil)
    (when (get-buffer rrse-process-buffer)
      (kill-buffer rrse-process-buffer))))

(defun rrse-query (type query)
  "Query to rrse"
  (rrse-init-process)
  (with-current-buffer (process-buffer rrse-process)
    (delete-region (point-min) (point-max))
    (process-send-string rrse-process (concat type " " query "\n"))
    (while (not (eq (char-before (point-max)) ?\n))
      (accept-process-output rrse-process 0 rrse-query-timeout))
    (let ((msg (buffer-substring-no-properties (point-min) (1- (point-max)))))
      (if (string= msg "") nil msg))))

(defun rrse-debug-message ()
  (concat (rrse-get-current-word)
          " "
          (if (rrse-method-call-p) "1" "0")
          " "
          (rrse-get-receiver)))

(defun rrse-build-minibuffer-help ()
  "Return help string for minibuffer from the method at the cursor"
  (let ((word (rrse-get-current-word))
        (receiver (rrse-get-receiver)))
    (if (rrse-method-call-p)
        (or (rrse-query "full-name" (concat receiver "." word))
            (rrse-query "instance-method" word))
      (rrse-query "toplevel-method" word))))

(defun rrse-print-current-symbol-info ()
  "Show the information of the method call at the cursor in the minibuffer"
  (interactive)
  (condition-case err
      (and rrse-enable
           (eq major-mode rrse-ruby-mode)
           (not executing-kbd-macro)
           (not (eq (selected-window) (minibuffer-window)))
           (sit-for rrse-idle-delay)
           (let ((minibuffer-help (rrse-build-minibuffer-help)))
             (when minibuffer-help (message minibuffer-help))))
    (error (princ err) (error "stop hook"))))

(defun rrse-toggle-show ()
  "Toggle whether documents are shown in minibuffer"
  (interactive)
  (setq rrse-enable (not rrse-enable)))

(defun rrse-setup ()
  "Setup rrse in emacs"
  (interactive)
  (add-hook rrse-ruby-mode-hook
            '(lambda ()
               (add-hook 'post-command-hook 'rrse-print-current-symbol-info))))

;;; rrse.el ends here

