require "rrse"
require "optparse"
require "bitclust"
require "msgpack"
#require "rdoc/ri/driver"

class RRSE::Command::UpdateDB
  def self.short_description
    "update database"
  end
  
  def initialize(dir)
    @dir = dir
    @table = {}
    @index_table = {}
  end

  def options
    opts = OptionParser.new
    opts.banner = "Usage: rrse [global-options] update-db"
    opts
  end

  def run(argv)
    options.order!(argv)
    config = RRSE.load_config(@dir)
    config["sources"].each do |source|
      case source["type"]
      when "bitclust"
        merge_tables(*BitClustAdaptor.new(
          read_bitclust_config(source["config"]), source["version"], source["dbpath"]
        ).generate_tables)
      when "ri"
        raise "ri does not support"
      end
    end

    save_table(@table, @dir.join("table"))
    save_table(@index_table, @dir.join("index_table"))
  end

  def read_bitclust_config(path)
    YAML.load_file(File.expand_path(path))
  end

  def merge_tables(table, index_table)
    @table.merge!(table)
    @index_table.merge!(index_table)
  end

  def save_table(table, path)
    File.open(path, "w"){|f| MessagePack.dump(table, f)}
  end
  
  class BitClustAdaptor
    def initialize(bitclust_config, version, dbpath)
      version ||= bitclust_config[:default_version]
      dbpath ||= bitclust_config[:database_prefix] + "-" + version
      @db = BitClust::MethodDatabase.new(dbpath)
    end

    def generate_tables
      all_entries = Hash.new{|h, k| h[k] = [] }
      index_table = Hash.new{|h, k| h[k] = [] }
      @db.classes.each do |c|
        c.entries.each do |m|
          typemark = normalize_typemark(m.typemark)
          next unless method_typemark?(typemark)
          m.names.each do |name|
            spec = "#{c.name}#{typemark}#{name}"
            all_entries[spec].concat(extract_params_from_source(name, m.source))
            index_table[name] << spec
          end
        end
      end
      return all_entries, index_table
    end

    def normalize_typemark(typemark)
      typemark == ".#" ? "." : typemark
    end

    def method_typemark?(typemark)
      typemark == "." || typemark == "#"
    end
    
    def extract_params_from_source(name, source)
      source.each_line.find_all{|line|
        /\A--- #{Regexp.quote(name)}[^a-zA-Z_\[\]=]/ =~ line
      }.map{|line|
        line.chomp.
          gsub(/\A--- #{Regexp.quote(name)}/, "").
          gsub(/\t/, "").
          gsub(/ +/, "").
          gsub(/->nil\Z/, "").
          gsub(/->\(\)\Z/, "")
      }
    end
  end

  class RIAdaptor
    def initialize(options)
      @ri = RDoc::RI::Driver.new(RDoc::RI::Driver.process_args(options))
      @ri.stores.each{|store| store.load_all }
    end

    def generate_tables
      @ri.stores.each{|store| p store.all_classes}
    end
  end
end



